# -*- coding: utf8 -*-
#
# cacaeditor    Libcaca ascii editor.
#
# Copyright (c) 2011 Alex Foulon <alxf@lavabit.com>
#               All Rights Reserved
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
#

""" Widgets for interface.
"""

#debug
#import pdb

#standard modules
import io
import os.path
from textwrap import wrap

#third party modules
import caca
from caca.canvas import Canvas, CanvasError, NullCanvas

#application module
import cacaeditor
from cacaeditor.helpers import ColorObject

#
# Base widgets
#

class CacaWidget(object):
    """ Object to define common methods for widgets with libcaca
        canvas attached.
    """
    def __init__(self, width, height):
        """ CacaWidget constructor.

            :width: the libcaca canvas width
            :height: the libcaca canvas height
        """
        self.init_caca(width, height)

    def __call__(self):
        """ Return libcaca canvas when instance is called.
        """
        return self.cv

    def init_caca(self, width, height=1):
        """ Create libcaca canvas.

            :width: the libcaca canvas width
            :height: the libcaca canvas height
        """
        try:
            self.cv = Canvas(width, height)
        except CanvasError, e:
            raise cacaeditor.AppError(e)

        self.make()

    def resize(self, width, height=1):
        """ Method to resize widget.

            :width:     the new width for widget
            :height:    the new height for widget
        """
        self.cv.set_size(width, height)
        self.make()

    def make(self):
        """ Overwrite this method to draw widget onto libcaca canvas.
        """
        raise cacaeditor.AppError("You need to overwrite 'make' method "
                                  "in your widget.")

class IndexSelect(object):
    """ Object to define common methods for selectable widgets in
        user interface like menus.
    """
    def select(self, increment=1):
        """ Method to switch to next selectable item.

            :increment: use a custom increment
        """
        self.idx += increment
        if self.idx > self.last_idx:
            self.idx = 0
        elif self.idx < 0:
            self.idx = self.last_idx

        self.make()

    def reset(self):
        """ Method to reset menu index.
        """
        self.idx = 0
        self.make()

class Document(object):
    """ Object to define common methods for document widgets.
    """
    def shortname(self):
        """ Return the basename for filename or "*scratch*" if no filename
            associated.
        """
        if self.filename is not None:
            return os.path.basename(self.filename)
        else:
            return "*scratch*"

    def load(self, filename, fmt=None):
        """ Load a filename into canvas.

            :filename: the file to load
            :format: format for filename to load
        """
        try:
            with io.open(filename, 'rb') as fp:
                try:
                    if fmt is None:
                        self.cv.import_from_memory(fp.read(), "")
                    else:
                        self.cv.import_from_memory(fp.read(), fmt)
                except CanvasError, err:
                    raise cacaeditor.AppError("Libcaca: %s" % err)
        except (IOError, OSError), err:
            raise cacaeditor.FileError(err)
        else:
            self.filename = filename
            self.data = self.cv.export_to_memory("caca")

    def save(self, filename=None):
        """ Export canvas to a file in caca format.

            :filename: the file to save
        """
        if filename is None and self.filename is None:
            return False

        try:
            if filename is not None:
                with io.open(filename, 'wb+') as fp:
                    fp.write(self.data)
                self.filename = filename
            elif self.filename is not None:
                with io.open(self.filename, 'wb+') as fp:
                    fp.write(self.data)
        except IOError, err:
            raise cacaeditor.FileError(err)
        else:
            return True

#
# Dialog widgets
#

class Dialog(CacaWidget):
    """ Model for dialogs.

        :message: the message to display

        keyword options:

            width       -- the dialog width (default: 50)
            wrap        -- wrap the text (default: True)
            w_size      -- the wrap size (default: 40)
            answer      -- if the dialog waiting for answer (default: False)
            title       -- the title of the dialog (default: None)
            callback    -- the callback for the dialog (default: None)
    """
    foreground = caca.COLOR_WHITE
    background = caca.COLOR_BLACK

    def __init__(self, message, **kw):
        """ Dialog constructor.
        """
        #set attributes
        self.msg = message
        self.width = 50
        self.wrap = True
        self.w_size = 40
        self.flag = False
        self.title = None
        self.callback = None

        #set keyword options
        if kw.has_key('width'):
            self.width = kw['width']
        if kw.has_key('wrap'):
            self.wrap = kw['wrap']
        if kw.has_key('w_size'):
            self.w_size = kw['w_size']
        if kw.has_key('answer'):
            self.flag = kw['answer']
        if kw.has_key('title'):
            self.title = kw['title']
        if kw.has_key('callback'):
            self.callback = kw['callback']

        #set message
        if self.wrap:
            self.msg = wrap(self.msg, self.w_size)
        else:
            self.msg = self.msg.split('\n')

        #format message
        height = len(self.msg) + 2
        if self.flag:
            self.answer = []
            height += 2

        #init caca
        self.init_caca(self.width, height)

    def make(self):
        """ Draw the dialog to canvas.
        """
        #reset libcaca canvas
        self.cv.clear()
        #init libcaca colors
        self.cv.set_color_ansi(self.foreground, self.background)

        #draw confirm box to canvas
        self.cv.fill_box(0, 0, self.cv.get_width(), self.cv.get_height(), ' ')
        self.cv.draw_cp437_box(0, 0, self.cv.get_width(), self.cv.get_height())
        if self.title is not None:
            self.cv.put_str(2, 0, " %s " % self.title)
        #print message
        x, y = 2, 1
        for line in self.msg:
            self.cv.put_str(x, y, line)
            y += 1

        if self.flag:
            x += 2
            y += 1
            self.cv.set_color_ansi(self.foreground, caca.COLOR_DARKGRAY)
            self.cv.draw_line(x, y, self.cv.get_width()-5, y, ' ')
            if self.answer:
                self.cv.put_str(x, y, "".join(self.answer))
            self.cv.set_color_ansi(self.foreground, self.background)

        #reset to default colors
        self.cv.set_color_ansi(caca.COLOR_DEFAULT, caca.COLOR_TRANSPARENT)

    def get_formxy(self, canvas):
        """ Return x, y cursor position for the form.

            :canvas: libcaca application canvas.
        """
        if self.flag:
            x = ((canvas.get_width() - self.cv.get_width()) / 2) + 4
            y = ((canvas.get_height() - self.cv.get_height()) / 2) + (self.cv.get_height() - 2)
            return (x, y)
        else:
            return (None, None)

class MsgDialog(Dialog):
    """ Build a dialog to display message.
    """
    foreground = caca.COLOR_WHITE
    background = caca.COLOR_BLUE

class MsgError(Dialog):
    """ Build a dialog to display error message.
    """
    foreground = caca.COLOR_BLACK
    background = caca.COLOR_RED

    def __init__(self, message, **kw):
        """ MsgError constructor.
        """
        kw['title'] = "Error"
        Dialog.__init__(self, message, **kw)

class MsgListBox(CacaWidget, IndexSelect):
    """ Build a list of choices.

        :message: a message to print
        :items: list of answers
    """
    foreground = caca.COLOR_WHITE
    background = caca.COLOR_BLUE

    def __init__(self, message, items, callback=None):
        """ MsgListBox constructor.
        """
        self.last_idx = len(items)
        self.idx = 1
        self.message = message
        self.items = items

        if callback is not None:
            self.callback = callback

        if len(message) > 50:
            width = len(message)
        else:
            width = 50

        height = len(self.items) + 4

        self.init_caca(width, height)

    def get_answer(self):
        """ Return selected item or None.
        """
        if self.idx == 0:
            return None
        else:
            return self.items[self.idx-1]

    def make(self):
        """ Draw the widget to canvas.
        """
        #reset canvas
        self.cv.clear()
        #set color for widget
        self.cv.set_color_ansi(self.foreground, self.background)

        #draw box
        self.cv.fill_box(0, 0, self.cv.get_width(), self.cv.get_height(), ' ')
        self.cv.draw_cp437_box(0, 0, self.cv.get_width(), self.cv.get_height())

        #draw message
        self.cv.put_str(2, 1, self.message)

        #draw choices list
        i, y = 1, 3
        for item in self.items:
            if self.idx == i:
                self.cv.set_color_ansi(self.background, self.foreground)
                self.cv.put_str(1, y, str(item).center(self.cv.get_width()-2))
                self.cv.set_color_ansi(self.foreground, self.background)
            else:
                self.cv.put_str(1, y, str(item).center(self.cv.get_width()-2))

            y += 1
            i += 1

        #reset to default colors
        self.cv.set_color_ansi(caca.COLOR_DEFAULT, caca.COLOR_TRANSPARENT)

    def select(self, increment=1):
        """ Method to switch to next selectable item.

            :increment: use a custom increment
        """
        #switch to next item
        self.idx += increment
        if self.idx > self.last_idx:
            self.idx = 1
        elif self.idx < 1:
            self.idx = self.last_idx

        #rebuild widget
        self.make()

#
# User interface widgets
#

class MenuBar(CacaWidget, IndexSelect):
    """ Build an horizontal menu.

        :width: the menu bar width
        :data:  data dictionnary for vertical submenu
    """
    #theme attributes
    foreground = caca.COLOR_WHITE
    background = caca.COLOR_BLUE

    def __init__(self, width, data=None):
        """ MenuBar constructor.
        """
        #set attributes
        self.subs = []
        self.data = data
        self.idx = 0
        self.last_idx = len(self.data['_index'])

        #add submenu foreach menu item
        if self.data is not None and isinstance(self.data, dict):
            for key in self.data['_index']:
                self.subs.append(SubMenu(self.data[key]))

        #intialize libcaca environment
        self.init_caca(width)

    def getx(self):
        """ Return x position for current selection.
        """
        x = 1

        #compute menu position
        if self.last_idx >= self.idx > 0:
            for i in range(1, self.idx):
                x += len(self.data['_index'][i-1]) + 2

        return x

    def make(self):
        """ Draw horizontal menu to canvas.
        """
        #reset libcaca canvas
        self.cv.clear()

        #set menu colors
        self.cv.set_color_ansi(self.foreground, self.background)

        #set background
        self.cv.draw_line(0, 0, self.cv.get_width(), 1, ' ')

        #build each menu item
        i, x = 1, 1
        for key in self.data['_index']:
            if self.idx == i:
                self.cv.set_color_ansi(self.background, self.foreground)
                self.cv.put_str(x, 0, " %s " % key)
                self.cv.set_color_ansi(self.foreground, self.background)
            else:
                self.cv.put_str(x, 0, " %s " % key)

            #increment counters
            i += 1
            x += len(key) + 2

        #reset to default colors
        self.cv.set_color_ansi(caca.COLOR_DEFAULT, caca.COLOR_TRANSPARENT)

    def reset(self):
        """ Reset menu and all submenus.
        """
        #reset menu index
        self.idx = 0

        #build menubar
        self.make()

        #reset all submenus
        for sub in self.subs:
            sub.reset()

class SubMenu(CacaWidget, IndexSelect):
    """ Build a vertical menu.
    """
    #theme attributes
    foreground = caca.COLOR_BLUE
    background = caca.COLOR_WHITE

    def __init__(self, items):
        """ SubMenu constructor.
        """
        #set attributes
        self.items = items
        self.idx = 0
        self.last_idx = len(items)

        #set max size for label + keybinding
        x, y = 0, 2
        for tag, desc, key in self.items:
            if tag is not None:
                if key is not None:
                    if x < (len(desc) + len(key) + 10):
                        x = (len(desc) + len(key) + 10)
                else:
                    if x < len(desc) + 10:
                        x = len(desc) + 10

            #increment height counter
            y += 1

        #initialize libcaca environment
        self.init_caca(x, y)

    def make(self):
        """ Draw the vertical menu to canvas
        """
        #reset canvas
        self.cv.clear()
        #init libcaca colors
        self.cv.set_color_ansi(self.foreground, self.background)

        #build vertical submenu to canvas
        self.cv.draw_cp437_box(0, 0, self.cv.get_width(), self.cv.get_height())
        y = 1
        for tag, desc, key in self.items:
            if tag is not None:
                if self.idx == y:
                    self.cv.set_color_ansi(self.background, self.foreground)
                    if key is not None:
                        self.cv.put_str(1, y, "%s%s%s" %\
                            (desc, "." * (self.cv.get_width() - \
                                (2 + len(desc) + len(key))), key))
                    else:
                        self.cv.put_str(1, y, "%s%s" %\
                            (desc, "." * (self.cv.get_width() - \
                                (2 + len(desc)))))
                    self.cv.set_color_ansi(self.foreground, self.background)
                else:
                    if key is not None:
                        self.cv.put_str(1, y, "%s%s%s" %\
                            (desc, "." * (self.cv.get_width() - \
                                (2 + len(desc) + len(key))), key))
                    else:
                        self.cv.put_str(1, y, "%s%s" %\
                            (desc, "." * (self.cv.get_width() - \
                                (2 + len(desc)))))
            else:
                self.cv.put_str(1, y, '-' * (self.cv.get_width() - 2))

            #increment counter
            y += 1

        #reset to default colors
        self.cv.set_color_ansi(caca.COLOR_DEFAULT, caca.COLOR_TRANSPARENT)

    def select(self, increment=1):
        """ Overwrite select method.
        """
        self.idx += increment
        if self.idx > self.last_idx:
            self.idx = 0
        elif self.idx < 0:
            self.idx = self.last_idx

        if self.items[self.idx-1][0] is None:
            self.select(increment)

        #build widget
        self.make()

class StatusBar(CacaWidget):
    """ Build a statusbar.
    """
    def __init__(self, width):
        """ StatusBar constructor.
        """
        self._left = None
        self._right = None

        self.init_caca(width)

    @apply
    def left():
        doc = "The left message attribute."

        def fget(self):
            return self._left

        def fset(self, value):
            self._left = value
            self.make()

        return property(**locals())

    @apply
    def right():
        doc = "The right message attribute."

        def fget(self):
            return self._right

        def fset(self, value):
            self._right = value
            self.make()

        return property(**locals())

    def make(self):
        """ Build the widget.
        """
        #reset canvas
        self.cv.clear()

        #init statusbar colors
        self.cv.set_color_ansi(caca.COLOR_WHITE, caca.COLOR_BLUE)

        #paint background
        self.cv.draw_line(0, 0, self.cv.get_width(), 1, ' ')

        #write left message if any (2/3 width)
        if self.left is not None:
            max_len = 2 * (self.cv.get_width() / 3)
            if len(self.left) > max_len:
                s = self.left[0:max_len-3] + "..."
            else:
                s = self.left
            self.cv.put_str(0, 0, s)

        #write right message if any (1/3 width)
        if self.right is not None:
            max_len = self.cv.get_width() / 3
            if len(self.right) > max_len:
                s = self.right[0:max_len-3] + "..."
            else:
                s = self.right
            x = self.cv.get_width() - (len(s)+1)
            self.cv.put_str(x, 0, s)

        #reset to default colors
        self.cv.set_color_ansi(caca.COLOR_DEFAULT, caca.COLOR_TRANSPARENT)

class Buffer(CacaWidget, Document):
    """ Build buffer widget.

        :width:     the buffer width
        :height:    the buffer height
        :filename:  file in caca format to open
    """
    def __init__(self, width, height, filename=None):
        """ Buffer constructor.
        """
        self.posx = 0
        self.posy = 0
        self.init_caca(width, height)
        self.colors = ColorObject(self.cv,
                                  caca.COLOR_DEFAULT, caca.COLOR_TRANSPARENT)

        self.filename = filename
        if self.filename is not None:
            self.load(self.filename, "caca")

    def make(self):
        """
        """
        pass

    def blit(self, x, y, cv):
        """ Blit a canvas into buffer.

            :x:     X coordinate for topleft corner
            :y:     Y coordinate for topleft corner
            :cv:    libcaca canvas to blit
        """
        self.cv.blit(x, y-1, cv, NullCanvas())
        self.data = self.cv.export_to_memory("caca")

    def insert_line(self, x1, y1, x2, y2, ch):
        """ Insert line / thin line into buffer document.

            :x1: X coordinate for first point
            :y1: Y coordinate for first point
            :x2: X coordinate for second point
            :y2: Y coordinate for second point
            :ch: character to be used to draw the line
        """
        if ch is None:
            self.cv.draw_thin_line(x1, y1-1, x2, y2-1)
        else:
            self.cv.draw_line(x1, y1-1, x2, y2-1, ch)

        self.data = self.cv.export_to_memory("caca")

    def insert_box(self, x, y, width, height, ch, fill=False):
        """ Insert box / thin box into buffer.

            :x:         X coordinate of the upper-left corner of the box
            :y:         Y coordinate of the upper-left corner of the box
            :width:     width of the box
            :height:    height of the box
            :ch:        character to be used to draw the box
            :fill:      indicate if the box must be filled with ch
        """
        if ch is None:
            self.cv.draw_thin_box(x, y-1, width+1, height+1)
        else:
            if fill:
                self.cv.fill_box(x, y-1, width+1, height+1, ch)
            else:
                self.cv.draw_box(x, y-1, width+1, height+1, ch)

        self.data = self.cv.export_to_memory("caca")

    def insert_triangle(self, x1, y1, x2, y2, x3, y3, ch, fill=False):
        """ Insert triangle / thin triangle into buffer.

            :x1:   X coordinate of the first point
            :y1:   Y coordinate of the first point
            :x2:   X coordinate of the second point
            :y2:   Y coordinate of the second point
            :x3:   X coordinate of the third point
            :y3:   Y coordinate of the third point
            :ch:   character to be used to draw the triangle outline
            :fill: indicate if the box must be filled with ch
        """
        if ch is None:
            self.cv.draw_thin_triangle(x1, y1-1, x2, y2-1, x3, y3-1)
        else:
            if fill:
                self.cv.fill_triangle(x1, y1-1, x2, y2-1, x3, y3-1, ch)
            else:
                self.cv.draw_triangle(x1, y1-1, x2, y2-1, x3, y3-1, ch)

        self.data = self.cv.export_to_memory("caca")

    def insert_circle(self, x, y, radius, ch):
        """ Insert circle into buffer.

            :x:         X coordinate of the center
            :y:         Y coordinate of the center
            :radius:    radius of the circle
            :ch:        character to be used to draw the circle outline
        """
        self.cv.draw_circle(x, y-1, radius, ch)

        self.data = self.cv.export_to_memory("caca")

    def insert_ellipse(self, x, y, ra, rb, ch, fill=False):
        """ Insert ellipse / thin ellipse into buffer.

            :x:     X coordinate of the center
            :y:     Y coordinate of the center
            :ra:    radius A for ellipse
            :rb:    radius B for ellipse
            :ch:    the character to be used to draw the ellipse outline
            :fill:  indicate if the box must be filled with ch
        """
        if ch is None:
            self.cv.draw_thin_ellipse(x, y-1, ra, rb)
        else:
            if fill:
                self.cv.fill_ellipse(x, y-1, ra, rb, ch)
            else:
                self.cv.draw_ellipse(x, y-1, ra, rb, ch)

        self.data = self.cv.export_to_memory("caca")

